@@ -25,11 +25,14 @@
 #include <gdk/gdkx.h>
 #include <unistd.h>
 #include <syslog.h>
+#include <grp.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <fcntl.h>
+#include <sys/stat.h>
 #include "gdm.h"
 #include "misc.h"
+#include "auth.h"
 
 #include <vicious.h>
 
@@ -40,6 +43,12 @@ extern char **stored_argv;
 extern int stored_argc;
 extern char *stored_path;
 
+/* Configuration option variables */
+extern gchar *GdmUser;
+extern gchar *GdmServAuthDir;
+extern uid_t GdmUserId;
+extern gid_t GdmGroupId;
+
 static int screenx = 0;
 static int screeny = 0;
 static int screenwidth = 0;
@@ -99,48 +108,68 @@ get_screen_size (GdmDisplay *d)
 static void
 center_window (GtkWidget *window)
 {
-	GtkRequisition req;
+	int w, h;
 
-	gtk_widget_size_request (window, &req);
+	/* sanity, should never happen */
+	if (window == NULL)
+		return;
+
+	gtk_window_get_size (GTK_WINDOW (window), &w, &h);
 
 	gtk_window_move (GTK_WINDOW (window),
 			 screenx +
 			 (screenwidth / 2) -
-			 (req.width / 2),
+			 (w / 2),
 			 screeny +
 			 (screenheight / 2) -
-			 (req.height / 2));
+			 (h / 2));
 }
 
 static void
 show_errors (GtkWidget *button, gpointer data)
 {
-	const char *file = data;
-	FILE *fp;
-	GtkWidget *sw;
-	GtkWidget *label;
-	GtkWidget *dlg = gtk_widget_get_toplevel (button);
-	GtkWidget *parent = button->parent;
-	GString *gs = g_string_new (NULL);
+	GtkRequisition req;
+	GtkWidget *textsw = data;
+	GtkWidget *dlg = g_object_get_data (G_OBJECT (button), "dlg");
 
-	fp = fopen (file, "r");
-	if (fp != NULL) {
-		char buf[256];
-		while (fgets (buf, sizeof (buf), fp))
-			g_string_append (gs, buf);
-		fclose (fp);
+	if (GTK_TOGGLE_BUTTON (button)->active) {
+		gtk_widget_show (textsw);
 	} else {
-		g_string_printf (gs, _("%s could not be opened"), file);
+		gtk_widget_hide (textsw);
 	}
 
-	gtk_widget_destroy (button);
+	/* keep window at the size request size */
+	gtk_widget_size_request (dlg, &req);
+	gtk_window_resize (GTK_WINDOW (dlg), req.width, req.height);
+}
+
+static GtkWidget *
+get_error_text_view (const char *details)
+{
+	GtkWidget *sw;
+	GtkWidget *tv;
+	GtkTextBuffer *buf;
+	GtkTextIter iter;
+
+	tv = gtk_text_view_new ();
+	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (tv));
+	gtk_text_view_set_editable (GTK_TEXT_VIEW (tv), FALSE);
+	gtk_text_buffer_create_tag (buf, "foo",
+				    "editable", FALSE,
+				    "family", "monospace",
+				    NULL);
+	gtk_text_buffer_get_iter_at_offset (buf, &iter, 0);
+
+	gtk_text_buffer_insert_with_tags_by_name
+		(buf, &iter,
+		 ve_sure_string (details), -1,
+		 "foo", NULL);
 
 	sw = gtk_scrolled_window_new (NULL, NULL);
 	if (gdk_screen_width () >= 800)
 		gtk_widget_set_size_request (sw, 500, 150);
 	else
 		gtk_widget_set_size_request (sw, 200, 150);
-	gtk_widget_show (sw);
 
 	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
 					GTK_POLICY_AUTOMATIC,
@@ -149,16 +178,58 @@ show_errors (GtkWidget *button, gpointer data)
 	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
 					     GTK_SHADOW_IN);
 
-	label = gtk_label_new (gs->str);
-	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw), label);
-	gtk_widget_show (label);
+	gtk_container_add (GTK_CONTAINER (sw), tv);
+	gtk_widget_show (tv);
 
-	gtk_box_pack_start (GTK_BOX (parent),
-			    sw, TRUE, TRUE, 0);
+	return sw;
+}
 
-	g_string_free (gs, TRUE);
+static void
+setup_dialog (GdmDisplay *d, const char *name, int closefdexcept)
+{
+	int argc = 1;
+	char **argv;
 
-	center_window (dlg);
+	closelog ();
+
+	gdm_close_all_descriptors (0 /* from */, closefdexcept /* except */, -1 /* except2 */);
+
+	/* No error checking here - if it's messed the best response
+	 * is to ignore & try to continue */
+	gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+	gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+	gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
+
+	setgid (GdmGroupId);
+	initgroups (GdmUser, GdmGroupId);
+	setuid (GdmUserId);
+
+	gdm_desetuid ();
+
+	/* restore initial environment */
+	gdm_restoreenv ();
+
+	openlog ("gdm", LOG_PID, LOG_DAEMON);
+
+	ve_setenv ("LOGNAME", GdmUser, TRUE);
+	ve_setenv ("USER", GdmUser, TRUE);
+	ve_setenv ("USERNAME", GdmUser, TRUE);
+
+	ve_setenv ("DISPLAY", d->name, TRUE);
+	ve_unsetenv ("XAUTHORITY");
+
+	gdm_auth_set_local_auth (d);
+
+	/* sanity env stuff */
+	ve_setenv ("SHELL", "/bin/sh", TRUE);
+	ve_setenv ("HOME", ve_sure_string (GdmServAuthDir), TRUE);
+
+	argv = g_new0 (char *, 2);
+	argv[0] = (char *)name;
+
+	gtk_init (&argc, &argv);
+
+	get_screen_size (d);
 }
 
 void
@@ -171,48 +242,65 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 
 	if (pid == 0) {
 		guint sid;
-		int argc = 1;
-		char **argv;
 		GtkWidget *dlg;
 		GtkWidget *button;
 		char *loc;
-		char *display;
-		char *xauthority;
+		char *details;
+		
+		/* First read the details if they exist */
+		if (details_file) {
+			FILE *fp;
+			struct stat s;
+			int r;
+			gboolean valid_utf8 = TRUE;
+			GString *gs = g_string_new (NULL);
 
-		closelog ();
+			fp = NULL;
+			IGNORE_EINTR (r = stat (details_file, &s));
+			if (r == 0) {
+				if (S_ISREG (s.st_mode))
+					fp = fopen (details_file, "r");
+				else {
+					loc = gdm_locale_to_utf8 (_("%s not a regular file!\n"));
+					g_string_printf (gs, loc, details_file);
+					g_free (loc);
+				}
+			}
+			if (fp != NULL) {
+				char buf[256];
+				int lines = 0;
+				while (fgets (buf, sizeof (buf), fp)) {
+					if ( ! g_utf8_validate (buf, -1, NULL))
+						valid_utf8 = FALSE;
+					g_string_append (gs, buf);
+					/* cap the lines at 500, that's already
+					   a possibility of 128k of crap */
+					if (lines ++ > 500) {
+						loc = gdm_locale_to_utf8 (_("\n... File too long to display ...\n"));
+						g_string_append (gs, loc);
+						g_free (loc);
+						break;
+					}
+				}
+				fclose (fp);
+			} else {
+				loc = gdm_locale_to_utf8 (_("%s could not be opened"));
+				g_string_append_printf (gs, loc, details_file);
+				g_free (loc);
+			}
 
-		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
+			details = g_string_free (gs, FALSE);
 
-		/* No error checking here - if it's messed the best response
-		 * is to ignore & try to continue */
-		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
-		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
-		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
+			if ( ! valid_utf8) {
+				char *tmp = gdm_locale_to_utf8 (details);
+				g_free (details);
+				details = tmp;
+			}
+		} else {
+			details = NULL;
+		}
 
-		gdm_desetuid ();
-
-		display = g_strdup (g_getenv ("DISPLAY"));
-		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
-
-		/* restore initial environment */
-		gdm_restoreenv ();
-
-		if (display != NULL)
-			gnome_setenv ("DISPLAY", display, TRUE);
-		if (xauthority != NULL)
-			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
-		/* sanity env stuff */
-		gnome_setenv ("SHELL", "/bin/bash", TRUE);
-		gnome_setenv ("HOME", "/", TRUE);
-
-		openlog ("gdm", LOG_PID, LOG_DAEMON);
-
-		argv = g_new0 (char *, 2);
-		argv[0] = "gtk-error-box";
-
-		gtk_init (&argc, &argv);
-
-		get_screen_size (d);
+		setup_dialog (d, "gtk-error-box", -1);
 
 		loc = gdm_locale_to_utf8 (error);
 
@@ -226,19 +314,26 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 
 		if (details_label != NULL) {
+			GtkWidget *text = get_error_text_view (details);
+
 			loc = gdm_locale_to_utf8 (details_label);
-			button = gtk_button_new_with_label (loc);
+			button = gtk_check_button_new_with_label (loc);
 			g_free (loc);
 
 			gtk_widget_show (button);
-			g_signal_connect (button, "clicked",
+			g_object_set_data (G_OBJECT (button), "dlg", dlg);
+			g_signal_connect (button, "toggled",
 					  G_CALLBACK (show_errors),
-					  /* leak? who cares we exit right
-					   * away */
-					  g_strdup (details_file));
+					  text);
 
 			gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox),
-					    button, FALSE, FALSE, 3);
+					    button, FALSE, FALSE, 6);
+			gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox),
+					    text, FALSE, FALSE, 6);
+
+			g_signal_connect_after (dlg, "size_allocate",
+						G_CALLBACK (center_window),
+						NULL);
 		}
 
 		button = gtk_dialog_add_button (GTK_DIALOG (dlg),
@@ -313,47 +408,10 @@ gdm_failsafe_question (GdmDisplay *d,
 	pid = gdm_fork_extra ();
 	if (pid == 0) {
 		guint sid;
-		int argc = 1;
-		char **argv;
 		GtkWidget *dlg, *label, *entry;
 		char *loc;
-		char *display;
-		char *xauthority;
 
-		closelog ();
-
-		gdm_close_all_descriptors (0 /* from */, p[1] /* except */);
-
-		/* No error checking here - if it's messed the best response
-		 * is to ignore & try to continue */
-		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
-		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
-		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
-
-		gdm_desetuid ();
-
-		display = g_strdup (g_getenv ("DISPLAY"));
-		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
-
-		/* restore initial environment */
-		gdm_restoreenv ();
-
-		if (display != NULL)
-			gnome_setenv ("DISPLAY", display, TRUE);
-		if (xauthority != NULL)
-			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
-		/* sanity env stuff */
-		gnome_setenv ("SHELL", "/bin/bash", TRUE);
-		gnome_setenv ("HOME", "/", TRUE);
-
-		openlog ("gdm", LOG_PID, LOG_DAEMON);
-
-		argv = g_new0 (char *, 2);
-		argv[0] = "gtk-failsafe-question";
-
-		gtk_init (&argc, &argv);
-
-		get_screen_size (d);
+		setup_dialog (d, "gtk-failsafe-question", p[1]);
 
 		loc = gdm_locale_to_utf8 (question);
 
@@ -423,17 +481,17 @@ gdm_failsafe_question (GdmDisplay *d,
 		char buf[BUFSIZ];
 		int bytes;
 
-		close (p[1]);
+		IGNORE_EINTR (close (p[1]));
 
 		gdm_wait_for_extra (NULL);
 
-		bytes = read (p[0], buf, BUFSIZ-1);
+		IGNORE_EINTR (bytes = read (p[0], buf, BUFSIZ-1));
 		if (bytes > 0) {
-			close (p[0]);
+			IGNORE_EINTR (close (p[0]));
 			buf[bytes] = '\0';
 			return g_strdup (buf);
 		} 
-		close (p[0]);
+		IGNORE_EINTR (close (p[0]));
 	} else {
 		gdm_error (_("%s: Cannot fork to display error/info box"),
 			   "gdm_failsafe_question");
@@ -454,47 +512,10 @@ gdm_failsafe_yesno (GdmDisplay *d,
 	pid = gdm_fork_extra ();
 	if (pid == 0) {
 		guint sid;
-		int argc = 1;
-		char **argv;
 		GtkWidget *dlg;
 		char *loc;
-		char *display;
-		char *xauthority;
 
-		closelog ();
-
-		gdm_close_all_descriptors (0 /* from */, p[1] /* except */);
-
-		/* No error checking here - if it's messed the best response
-		 * is to ignore & try to continue */
-		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
-		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
-		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
-
-		gdm_desetuid ();
-
-		display = g_strdup (g_getenv ("DISPLAY"));
-		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
-
-		/* restore initial environment */
-		gdm_restoreenv ();
-
-		if (display != NULL)
-			gnome_setenv ("DISPLAY", display, TRUE);
-		if (xauthority != NULL)
-			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
-		/* sanity env stuff */
-		gnome_setenv ("SHELL", "/bin/bash", TRUE);
-		gnome_setenv ("HOME", "/", TRUE);
-
-		openlog ("gdm", LOG_PID, LOG_DAEMON);
-
-		argv = g_new0 (char *, 2);
-		argv[0] = "gtk-failsafe-yesno";
-
-		gtk_init (&argc, &argv);
-
-		get_screen_size (d);
+		setup_dialog (d, "gtk-failsafe-yesno", p[1]);
 
 		loc = gdm_locale_to_utf8 (question);
 
@@ -531,9 +552,9 @@ gdm_failsafe_yesno (GdmDisplay *d,
 		setup_cursor (GDK_LEFT_PTR);
 
 		if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_YES)
-			gdm_fdprintf (STDOUT_FILENO, "yes\n");
+			gdm_fdprintf (p[1], "yes\n");
 		else
-			gdm_fdprintf (STDOUT_FILENO, "no\n");
+			gdm_fdprintf (p[1], "no\n");
 
 		XSetInputFocus (GDK_DISPLAY (),
 				PointerRoot,
@@ -545,24 +566,121 @@ gdm_failsafe_yesno (GdmDisplay *d,
 		char buf[BUFSIZ];
 		int bytes;
 
-		close (p[1]);
+		IGNORE_EINTR (close (p[1]));
 
 		gdm_wait_for_extra (NULL);
 
-		bytes = read (p[0], buf, BUFSIZ-1);
+		IGNORE_EINTR (bytes = read (p[0], buf, BUFSIZ-1));
 		if (bytes > 0) {
-			close (p[0]);
+			IGNORE_EINTR (close (p[0]));
 			if (buf[0] == 'y')
 				return TRUE;
 			else
 				return FALSE;
 		} 
-		close (p[0]);
+		IGNORE_EINTR (close (p[0]));
 	} else {
 		gdm_error (_("%s: Cannot fork to display error/info box"),
-			   "gdm_failsafe_question");
+			   "gdm_failsafe_yesno");
 	}
 	return FALSE;
 }
 
+int
+gdm_failsafe_ask_buttons (GdmDisplay *d,
+			  const char *question,
+			  char **but)
+{
+	pid_t pid;
+	int p[2];
+
+	if (pipe (p) < 0)
+		return -1;
+
+	pid = gdm_fork_extra ();
+	if (pid == 0) {
+		int i;
+		guint sid;
+		GtkWidget *dlg;
+		char *loc;
+
+		setup_dialog (d, "gtk-failsafe-ask-buttons", p[1]);
+
+		loc = gdm_locale_to_utf8 (question);
+
+		dlg = gtk_message_dialog_new (NULL /* parent */,
+					      0 /* flags */,
+					      GTK_MESSAGE_QUESTION,
+					      GTK_BUTTONS_NONE,
+					      "%s",
+					      loc);
+		g_free (loc);
+		for (i = 0; but[i] != NULL; i++) {
+			loc = gdm_locale_to_utf8 (but[i]);
+			gtk_dialog_add_button (GTK_DIALOG (dlg),
+					       loc, i);
+			g_free (loc);
+
+		}
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
+
+		sid = g_signal_lookup ("event",
+				       GTK_TYPE_WIDGET);
+		g_signal_add_emission_hook (sid,
+					    0 /* detail */,
+					    gdm_event,
+					    NULL /* data */,
+					    NULL /* destroy_notify */);
+
+		center_window (dlg);
+
+		gtk_widget_show_now (dlg);
+
+		if (dlg->window != NULL) {
+			gdk_error_trap_push ();
+			XSetInputFocus (GDK_DISPLAY (),
+					GDK_WINDOW_XWINDOW (dlg->window),
+					RevertToPointerRoot,
+					CurrentTime);
+			gdk_flush ();
+			gdk_error_trap_pop ();
+		}
+
+		setup_cursor (GDK_LEFT_PTR);
+
+		i = gtk_dialog_run (GTK_DIALOG (dlg));
+		gdm_fdprintf (p[1], "%d\n", i);
+
+		XSetInputFocus (GDK_DISPLAY (),
+				PointerRoot,
+				RevertToPointerRoot,
+				CurrentTime);
+
+		_exit (0);
+	} else if (pid > 0) {
+		char buf[BUFSIZ];
+		int bytes;
+
+		IGNORE_EINTR (close (p[1]));
+
+		gdm_wait_for_extra (NULL);
+
+		IGNORE_EINTR (bytes = read (p[0], buf, BUFSIZ-1));
+		if (bytes > 0) {
+			int i;
+			IGNORE_EINTR (close (p[0]));
+			buf[bytes] = '\0';
+			if (sscanf (buf, "%d", &i) == 1)
+				return i;
+			else
+				return -1;
+		} 
+		IGNORE_EINTR (close (p[0]));
+	} else {
+		gdm_error (_("%s: Cannot fork to display error/info box"),
+			   "gdm_failsafe_ask_buttons");
+	}
+	return -1;
+}
+
 /* EOF */
