@@ -20,6 +20,7 @@
 #include <libgnome/libgnome.h>
 #include <syslog.h>
 #include <unistd.h>
+#include <dirent.h>
 #include <signal.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -63,6 +64,8 @@ extern pid_t gdm_main_pid;
 extern gboolean preserve_ld_vars;
 extern int gdm_in_signal;
 
+extern gboolean no_console;
+
 extern char *gdm_charset;
 
 static void
@@ -217,7 +220,7 @@ gdm_fdprintf (int fd, const gchar *format, ...)
 	s = g_strdup_vprintf (format, args);
 	va_end (args);
 
-	write (fd, s, strlen (s));
+	IGNORE_EINTR (write (fd, s, strlen (s)));
 
 	g_free (s);
 }
@@ -324,6 +327,7 @@ gdm_get_free_display (int start, uid_t server_uid)
 		struct stat s;
 		char buf[256];
 		FILE *fp;
+		int r;
 
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *dsp = li->data;
@@ -341,17 +345,19 @@ gdm_get_free_display (int start, uid_t server_uid)
 		serv_addr.sin_port = htons (6000 + i);
 
 		errno = 0;
-		if (connect (sock, (struct sockaddr *)&serv_addr,
-			     sizeof (serv_addr)) >= 0 ||
-		    errno != ECONNREFUSED) {
-			close (sock);
+		IGNORE_EINTR (connect (sock,
+				       (struct sockaddr *)&serv_addr,
+				       sizeof (serv_addr)));
+		if (errno != 0 && errno != ECONNREFUSED) {
+			IGNORE_EINTR (close (sock));
 			continue;
 		}
-		close (sock);
+		IGNORE_EINTR (close (sock));
 
 		/* if lock file exists and the process exists */
 		g_snprintf (buf, sizeof (buf), "/tmp/.X%d-lock", i);
-		if (stat (buf, &s) == 0 &&
+		IGNORE_EINTR (r = stat (buf, &s));
+		if (r == 0 &&
 		    ! S_ISREG (s.st_mode)) {
 			/* Eeeek! not a regular file?  Perhaps someone
 			   is trying to play tricks on us */
@@ -372,7 +378,7 @@ gdm_get_free_display (int start, uid_t server_uid)
 			fclose (fp);
 
 			/* whack the file, it's a stale lock file */
-			unlink (buf);
+			IGNORE_EINTR (unlink (buf));
 		}
 
 		/* if starting as root, we'll be able to overwrite any
@@ -381,14 +387,16 @@ gdm_get_free_display (int start, uid_t server_uid)
 		if (server_uid > 0) {
 			g_snprintf (buf, sizeof (buf),
 				    "/tmp/.X11-unix/X%d", i);
-			if (stat (buf, &s) == 0 &&
+			IGNORE_EINTR (r = stat (buf, &s));
+			if (r == 0 &&
 			    s.st_uid != server_uid) {
 				continue;
 			}
 
 			g_snprintf (buf, sizeof (buf),
 				    "/tmp/.X%d-lock", i);
-			if (stat (buf, &s) == 0 &&
+			IGNORE_EINTR (r = stat (buf, &s));
+			if (r == 0 &&
 			    s.st_uid != server_uid) {
 				continue;
 			}
@@ -406,6 +414,9 @@ gdm_text_message_dialog (const char *msg)
 	char *dialog; /* do we have dialog?*/
 	char *msg_quoted;
 
+	if (no_console)
+		return FALSE;
+
 	if (access (EXPANDED_LIBEXECDIR "/gdmopen", X_OK) != 0)
 		return FALSE;
 
@@ -432,10 +443,12 @@ gdm_text_message_dialog (const char *msg)
 				   TRUE /* de_setuid */) < 0) {
 			g_free (dialog);
 			g_free (msg_quoted);
+			g_free (argv[4]);
 			return FALSE;
 		}
 
 		g_free (dialog);
+		g_free (argv[4]);
 	} else {
 		char *argv[6];
 
@@ -451,11 +464,11 @@ gdm_text_message_dialog (const char *msg)
 
 		if (gdm_exec_wait (argv, TRUE /* no display */,
 				   TRUE /* de_setuid */) < 0) {
-			g_free (argv[3]);
+			g_free (argv[4]);
 			g_free (msg_quoted);
 			return FALSE;
 		}
-		g_free (argv[3]);
+		g_free (argv[4]);
 	}
 	g_free (msg_quoted);
 	return TRUE;
@@ -466,6 +479,9 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 {
 	char *dialog; /* do we have dialog?*/
 	char *msg_quoted;
+
+	if (no_console)
+		return FALSE;
 	
 	if (access (EXPANDED_LIBEXECDIR "/gdmopen", X_OK) != 0)
 		return FALSE;
@@ -497,6 +513,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 		retint = gdm_exec_wait (argv, TRUE /* no display */,
 					TRUE /* de_setuid */);
 		if (retint < 0) {
+			g_free (argv[4]);
 			g_free (dialog);
 			g_free (msg_quoted);
 			return FALSE;
@@ -507,6 +524,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 
 		g_free (dialog);
 		g_free (msg_quoted);
+		g_free (argv[4]);
 
 		return TRUE;
 	} else {
@@ -522,7 +540,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 			return FALSE;
 		}
 
-		close (tempfd);
+		IGNORE_EINTR (close (tempfd));
 
 		argv[0] = EXPANDED_LIBEXECDIR "/gdmopen";
 		argv[1] = "-l";
@@ -540,11 +558,11 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 
 		if (gdm_exec_wait (argv, TRUE /* no display */,
 				   TRUE /* de_setuid */) < 0) {
-			g_free (argv[3]);
+			g_free (argv[4]);
 			g_free (msg_quoted);
 			return FALSE;
 		}
-		g_free (argv[3]);
+		g_free (argv[4]);
 
 		if (ret != NULL) {
 			fp = fopen (tempname, "r");
@@ -559,7 +577,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 			}
 		}
 
-		unlink (tempname);
+		IGNORE_EINTR (unlink (tempname));
 
 		g_free (msg_quoted);
 		return TRUE;
@@ -601,7 +619,7 @@ gdm_exec_wait (char * const *argv, gboolean no_display,
 			ve_unsetenv ("XAUTHORITY");
 		}
 		
-		execv (argv[0], argv);
+		IGNORE_EINTR (execv (argv[0], argv));
 
 		_exit (-1);
 	}
@@ -745,7 +763,7 @@ gdm_wait_for_extra (int *status)
 {
 	gdm_sigchld_block_push ();
 
-	if (extra_process > 0) {
+	if (extra_process > 1) {
 		ve_waitpid_no_signal (extra_process, &extra_status, 0);
 	}
 	extra_process = 0;
@@ -771,15 +789,18 @@ gdm_ensure_sanity (void)
 
         if (mkdir ("/tmp/.ICE-unix", 0777) == 0) {
 		/* Make sure it is root */
-		if (chown ("/tmp/.ICE-unix", 0, 0) == 0)
-			chmod ("/tmp/.ICE-unix", 01777);
+		IGNORE_EINTR (chown ("/tmp/.ICE-unix", 0, 0));
+		if (errno != 0)
+			IGNORE_EINTR (chmod ("/tmp/.ICE-unix", 01777));
         } else {
 		struct stat s;
-		if (lstat ("/tmp/.ICE-unix", &s) == 0 &&
-		    S_ISDIR (s.st_mode)) {
+		int r;
+		IGNORE_EINTR (r = lstat ("/tmp/.ICE-unix", &s));
+		if (r == 0 && S_ISDIR (s.st_mode)) {
 			/* Make sure it is root and sticky */
-			if (chown ("/tmp/.ICE-unix", 0, 0) == 0)
-				chmod ("/tmp/.ICE-unix", 01777);
+			IGNORE_EINTR (chown ("/tmp/.ICE-unix", 0, 0));
+			if (errno != 0)
+				IGNORE_EINTR (chmod ("/tmp/.ICE-unix", 01777));
 		}
 	}
 
@@ -837,7 +858,7 @@ gdm_peek_local_address_list (void)
 			   "gdm_peek_local_address_list");
 		g_free (buf);
 		if (sockfd != gdm_xdmcpfd)
-			close (sockfd);
+			IGNORE_EINTR (close (sockfd));
 		addr = g_new0 (struct in_addr, 1);
 		addr->s_addr = INADDR_LOOPBACK;
 		return g_list_append (NULL, addr);
@@ -869,7 +890,7 @@ gdm_peek_local_address_list (void)
 	}
 
 	if (sockfd != gdm_xdmcpfd)
-		close (sockfd);
+		IGNORE_EINTR (close (sockfd));
 
 	g_free (buf);
 #else /* SIOCGIFCONF */
@@ -1058,7 +1079,7 @@ gdm_fdgetc (int fd)
 	char buf[1];
 	int bytes;
 
-	bytes = read (fd, buf, 1);
+	IGNORE_EINTR (bytes = read (fd, buf, 1));
 	if (bytes != 1)
 		return EOF;
 	else
@@ -1093,11 +1114,45 @@ gdm_fdgets (int fd)
 void
 gdm_close_all_descriptors (int from, int except, int except2)
 {
-	int i;
-	int max = sysconf (_SC_OPEN_MAX);
-	for (i = from; i < max; i++) {
-		if (i != except && i != except2)
-			close(i);
+	DIR *dir;
+	struct dirent *ent;
+	GSList *openfds = NULL;
+
+	/* evil, but less evil then going to _SC_OPEN_MAX
+	   which can be very VERY large */
+	dir = opendir ("/proc/self/fd/");
+	if (dir != NULL) {
+		GSList *li;
+		while ((ent = readdir (dir)) != NULL) {
+			openfds = g_slist_prepend (openfds,
+						   GINT_TO_POINTER (atoi (ent->d_name)));
+		}
+		closedir (dir);
+		for (li = openfds; li != NULL; li = li->next) {
+			int fd = GPOINTER_TO_INT (li->data); 
+			if (fd >= from && fd != except && fd != except2) {
+				IGNORE_EINTR (close(fd));
+			}
+		}
+		g_slist_free (openfds);
+	} else {
+		int i;
+		int max = sysconf (_SC_OPEN_MAX);
+		/* don't go higher then this.  This is
+		 * a safety measure to not hang on crazy
+		 * systems */
+		if (max > 4096) {
+			/* FIXME: warn about this perhaps */
+			/* try an open, in case we're really
+			   leaking fds somewhere badly, this
+			   should be very high */
+			i = gdm_open_dev_null (O_RDONLY);
+			max = MAX (i+1, 4096);
+		}
+		for (i = from; i < max; i++) {
+			if (i != except && i != except2)
+				IGNORE_EINTR (close(i));
+		}
 	}
 }
 
@@ -1121,15 +1176,45 @@ gdm_unset_signals (void)
 	sigemptyset (&mask);
 	sigprocmask (SIG_SETMASK, &mask, NULL);
 
-	signal (SIGUSR1, SIG_DFL);
-	signal (SIGUSR2, SIG_DFL);
-	signal (SIGCHLD, SIG_DFL);
-	signal (SIGTERM, SIG_DFL);
-	signal (SIGPIPE, SIG_DFL);
-	signal (SIGALRM, SIG_DFL);
-	signal (SIGHUP, SIG_DFL);
+	gdm_signal_default (SIGUSR1);
+	gdm_signal_default (SIGUSR2);
+	gdm_signal_default (SIGCHLD);
+	gdm_signal_default (SIGTERM);
+	gdm_signal_default (SIGINT);
+	gdm_signal_default (SIGPIPE);
+	gdm_signal_default (SIGALRM);
+	gdm_signal_default (SIGHUP);
 }
 
+void
+gdm_signal_ignore (int signal)
+{
+	struct sigaction ign_signal;
+
+	ign_signal.sa_handler = SIG_IGN;
+	ign_signal.sa_flags = SA_RESTART;
+	sigemptyset (&ign_signal.sa_mask);
+
+	if (sigaction (signal, &ign_signal, NULL) < 0)
+		gdm_error (_("%s: Error setting signal %d to %s"),
+			   "gdm_signal_ignore", signal, "SIG_IGN");
+}
+
+void
+gdm_signal_default (int signal)
+{
+	struct sigaction def_signal;
+
+	def_signal.sa_handler = SIG_DFL;
+	def_signal.sa_flags = SA_RESTART;
+	sigemptyset (&def_signal.sa_mask);
+
+	if (sigaction (signal, &def_signal, NULL) < 0)
+		gdm_error (_("%s: Error setting signal %d to %s"),
+			   "gdm_signal_ignore", signal, "SIG_DFL");
+}
+
+
 static char *
 ascify (const char *text)
 {
@@ -1240,6 +1325,13 @@ static struct sigaction oldterm, oldint, oldhup;
 static void
 jumpback_sighandler (int signal)
 {
+	/* this avoids a race see Note below.
+	   We want to jump back only on the first
+	   signal invocation, even if the signal
+	   handler didn't return. */
+	gboolean old_do_jumpback = do_jumpback;
+	do_jumpback = FALSE;
+
 	if (signal == SIGINT)
 		oldint.sa_handler (signal);
 	else if (signal == SIGTERM)
@@ -1247,13 +1339,24 @@ jumpback_sighandler (int signal)
 	else if (signal == SIGHUP)
 		oldint.sa_handler (signal);
 	/* no others should be set up */
+
+	/* Note that we may not get here since
+	   the SIGTERM handler in slave.c
+	   might have in fact done the big Longjmp
+	   to the slave's death */
 	
-	if (do_jumpback) {
-		do_jumpback = FALSE;
+	if (old_do_jumpback) {
 		Longjmp (signal_jumpback, 1);
 	}
 }
 
+/*
+ * This sets up interruptes to be proxied and the
+ * gethostbyname/addr to be whacked using longjmp,
+ * in case INT/TERM/HUP was gotten in which case
+ * we no longer care for the result of the
+ * resolution.
+ */
 #define SETUP_INTERRUPTS_FOR_TERM_DECLS \
     struct sigaction term;
 
@@ -1265,36 +1368,37 @@ jumpback_sighandler (int signal)
     sigemptyset (&term.sa_mask);					\
 									\
     if (sigaction (SIGTERM, &term, &oldterm) < 0) 			\
-	gdm_fail (_("%s: Error setting up TERM signal handler"),	\
-		  "SETUP_INTERRUPTS_FOR_TERM");				\
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "TERM", strerror (errno)); \
 									\
     if (sigaction (SIGINT, &term, &oldint) < 0)				\
-	gdm_fail (_("%s: Error setting up INT signal handler"),		\
-		  "SETUP_INTERRUPTS_FOR_TERM");				\
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "INT", strerror (errno)); \
 									\
     if (sigaction (SIGHUP, &term, &oldhup) < 0) 			\
-	gdm_fail (_("%s: Error setting up HUP signal handler"),		\
-		  "SETUP_INTERRUPTS_FOR_TERM");
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "HUP", strerror (errno)); \
 
 #define SETUP_INTERRUPTS_FOR_TERM_TEARDOWN \
     do_jumpback = FALSE;						\
 									\
     if (sigaction (SIGTERM, &oldterm, NULL) < 0) 			\
-	gdm_fail (_("%s: Error setting up TERM signal handler"),	\
-		  "SETUP_INTERRUPTS_FOR_TERM");				\
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "TERM", strerror (errno)); \
 									\
     if (sigaction (SIGINT, &oldint, NULL) < 0) 				\
-	gdm_fail (_("%s: Error setting up INT signal handler"),		\
-		  "SETUP_INTERRUPTS_FOR_TERM");				\
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "INT", strerror (errno)); \
 									\
     if (sigaction (SIGHUP, &oldhup, NULL) < 0) 				\
-	gdm_fail (_("%s: Error setting up HUP signal handler"),		\
-		  "SETUP_INTERRUPTS_FOR_TERM");
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),	\
+		  "SETUP_INTERRUPTS_FOR_TERM", "HUP", strerror (errno));
 
 GdmHostent *
 gdm_gethostbyname (const char *name)
 {
-	struct hostent *he_;
+	/* static because of Setjmp */
+	static struct hostent * he_;
 	SETUP_INTERRUPTS_FOR_TERM_DECLS
 
 	/* the cached address */
@@ -1302,6 +1406,8 @@ gdm_gethostbyname (const char *name)
 	static time_t last_time = 0;
 	static char *cached_hostname = NULL;
 
+	he_ = NULL;
+
 	if (cached_hostname != NULL &&
 	    strcmp (cached_hostname, name) == 0) {
 		/* don't check more then every 60 seconds */
@@ -1336,7 +1442,8 @@ gdm_gethostbyname (const char *name)
 GdmHostent *
 gdm_gethostbyaddr (struct in_addr *ia)
 {
-	struct hostent *he_;
+	/* static because of Setjmp */
+	static struct hostent * he_;
 	SETUP_INTERRUPTS_FOR_TERM_DECLS
 
 	/* the cached address */
@@ -1344,6 +1451,8 @@ gdm_gethostbyaddr (struct in_addr *ia)
 	static time_t last_time = 0;
 	static struct in_addr cached_addr;
 
+	he_ = NULL;
+
 	if (last_time != 0 &&
 	    memcmp (&cached_addr, ia, sizeof (struct in_addr)) == 0) {
 		/* don't check more then every 60 seconds */
@@ -1416,7 +1525,7 @@ FILE *
 gdm_safe_fopen_w (const char *file)
 {
 	int fd;
-	unlink (file);
+	IGNORE_EINTR (unlink (file));
 	fd = open (file, O_EXCL|O_CREAT|O_TRUNC|O_WRONLY, 0644);
 	if (fd < 0)
 		return NULL;
@@ -1513,4 +1622,20 @@ gdm_root_user (void)
 	return root_user;
 }
 
+void
+gdm_sleep_no_signal (int secs)
+{
+	time_t endtime = time (NULL)+secs;
+
+	while (secs > 0) {
+		struct timeval tv;
+		tv.tv_sec = secs;
+		tv.tv_usec = 0;
+		select (0, NULL, NULL, NULL, &tv);
+		/* don't want to use sleep since we're using alarm
+		   for pinging */
+		secs = endtime - time (NULL);
+	}
+}
+
 /* EOF */
