@@ -47,6 +47,8 @@ extern gchar *GdmUser;
 extern gboolean GdmAllowRemoteAutoLogin;
 extern gint GdmRetryDelay;
 
+extern gboolean no_console;
+
 /* Evil, but this way these things are passed to the child session */
 static pam_handle_t *pamh = NULL;
 
@@ -60,6 +62,9 @@ static gboolean did_we_ask_for_password = FALSE;
 
 static char *selected_user = NULL;
 
+static gboolean opened_session = FALSE;
+static gboolean did_setcred = FALSE;
+
 void
 gdm_verify_select_user (const char *user)
 {
@@ -122,10 +127,13 @@ gdm_verify_pam_conv (int num_msg, const struct pam_message **msg,
     char *s;
     struct pam_response *reply = NULL;
     const char *login;
+
+    if (pamh == NULL)
+	return PAM_CONV_ERR;
     
     reply = malloc (sizeof (struct pam_response) * num_msg);
     
-    if (!reply) 
+    if (reply == NULL)
 	return PAM_CONV_ERR;
 
     memset (reply, 0, sizeof (struct pam_response) * num_msg);
@@ -140,6 +148,10 @@ gdm_verify_pam_conv (int num_msg, const struct pam_message **msg,
 		    login = tmp_PAM_USER;
 	    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
     }
+
+    /* Workaround to avoid gdm messages being logged as PAM_pwdb */
+    closelog ();
+    openlog ("gdm", LOG_PID, LOG_DAEMON);
     
     for (replies = 0; replies < num_msg; replies++) {
 	gboolean islogin = FALSE;
@@ -261,9 +273,12 @@ gdm_verify_standalone_pam_conv (int num_msg, const struct pam_message **msg,
 	char *s, *text;
 	struct pam_response *reply = NULL;
 
+	if (pamh == NULL)
+		return PAM_CONV_ERR;
+    
 	reply = malloc (sizeof (struct pam_response) * num_msg);
 
-	if (!reply) 
+	if (reply == NULL)
 		return PAM_CONV_ERR;
 
 	memset (reply, 0, sizeof (struct pam_response) * num_msg);
@@ -366,10 +381,14 @@ create_pamh (GdmDisplay *d,
 		gdm_error ("create_pamh: Stale pamh around, cleaning up");
 		pam_end (pamh, PAM_SUCCESS);
 	}
+	/* init things */
 	pamh = NULL;
+	opened_session = FALSE;
+	did_setcred = FALSE;
 
 	/* Initialize a PAM session for the user */
 	if ((*pamerr = pam_start (service, login, conv, &pamh)) != PAM_SUCCESS) {
+		pamh = NULL; /* be anal */
 		if (gdm_slave_should_complain ())
 			gdm_error (_("Unable to establish service %s: %s\n"),
 				   service, pam_strerror (NULL, *pamerr));
@@ -383,17 +402,6 @@ create_pamh (GdmDisplay *d,
 		return FALSE;
 	}
 
-#if 0
-	/* Apparently we should not set RUSER, docs are unclear but the sun
-           guys are saying don't do it */
-	/* gdm is requesting the login */
-	if ((*pamerr = pam_set_item (pamh, PAM_RUSER, GdmUser)) != PAM_SUCCESS) {
-		if (gdm_slave_should_complain ())
-			gdm_error (_("Can't set PAM_RUSER=%s"), GdmUser);
-		return FALSE;
-	}
-#endif
-
 	if ( ! d->console) {
 		/* Only set RHOST if host is remote */
 		/* From the host of the display */
@@ -450,15 +458,6 @@ gdm_verify_user (GdmDisplay *d,
 
     cur_gdm_disp = d;
 
-    /* Initialize a PAM session for the user */
-    if ( ! create_pamh (d, "gdm", username, &pamc, display, &pamerr)) {
-	    if (started_timer)
-		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
-	    goto pamerr;
-    }
-
-    pam_set_item (pamh, PAM_USER_PROMPT, _("Username:"));
-
     /* A Solaris thing: */
 #ifdef HAVE_DEFOPEN
     if (defopen(DEFLT"/login") == 0) {
@@ -477,33 +476,63 @@ gdm_verify_user (GdmDisplay *d,
     }
 #endif
 	    
-#ifdef PAM_FAIL_DELAY
-    pam_fail_delay (pamh, GdmRetryDelay * 1000);
-#endif /* PAM_FAIL_DELAY */
-
 authenticate_again:
+
     /* hack */
     g_free (tmp_PAM_USER);
     tmp_PAM_USER = NULL;
 
+    /* Initialize a PAM session for the user */
+    if ( ! create_pamh (d, "gdm", login, &pamc, display, &pamerr)) {
+	    if (started_timer)
+		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
+	    goto pamerr;
+    }
+
+    pam_set_item (pamh, PAM_USER_PROMPT, _("Username:"));
+
+#ifdef PAM_FAIL_DELAY
+    pam_fail_delay (pamh, GdmRetryDelay * 1000);
+#endif /* PAM_FAIL_DELAY */
+
     did_we_ask_for_password = FALSE;
 
     gdm_verify_select_user (NULL);
     /* Start authentication session */
     if ((pamerr = pam_authenticate (pamh, null_tok)) != PAM_SUCCESS) {
 	    if ( ! ve_string_empty (selected_user)) {
+		    pam_handle_t *tmp_pamh;
+
+		    /* Face browser was used to select a user,
+		       just completely rewhack everything since it
+		       seems various PAM implementations are
+		       having goats with just setting PAM_USER
+		       and trying to pam_authenticate again */
+
+		    g_free (login);
+		    login = selected_user;
+		    selected_user = NULL;
+
+		    gdm_sigterm_block_push ();
+		    gdm_sigchld_block_push ();
+		    tmp_pamh = pamh;
+		    pamh = NULL;
+		    gdm_sigchld_block_pop ();
+		    gdm_sigterm_block_pop ();
+
 		    /* FIXME: what about errors */
-		    pam_set_item (pamh, PAM_USER, selected_user);
-		    /* Note that the GDM_SETUSER will be sent in the
-		       authenticate conversation.  This is a more robust
-		       solution. */
+		    /* really this has been a sucess, not a failure */
+		    pam_end (tmp_pamh, PAM_SUCCESS);
+
+		    gdm_slave_greeter_ctl_no_ret (GDM_SETLOGIN, login);
+
 		    goto authenticate_again;
 	    }
 	    if (started_timer)
 		    gdm_slave_greeter_ctl_no_ret (GDM_STOPTIMER, "");
 	    if (gdm_slave_should_complain ()) {
 #ifndef PAM_FAIL_DELAY
-		    sleep (GdmRetryDelay);
+		    gdm_sleep_no_signal (GdmRetryDelay);
 #endif /* PAM_FAIL_DELAY */
 		    gdm_error (_("Couldn't authenticate user"));
 	    }
@@ -605,9 +634,11 @@ authenticate_again:
 	    goto pamerr;
     }
 
+    did_setcred = TRUE;
     /* Set credentials */
     pamerr = pam_setcred (pamh, PAM_ESTABLISH_CRED);
     if (pamerr != PAM_SUCCESS) {
+        did_setcred = FALSE;
 	if (gdm_slave_should_complain ())
 	    gdm_error (_("Couldn't set credentials for %s"), login);
 	goto pamerr;
@@ -615,9 +646,13 @@ authenticate_again:
 
     credentials_set = TRUE;
 
+    opened_session = TRUE;
     /* Register the session */
     pamerr = pam_open_session (pamh, 0);
     if (pamerr != PAM_SUCCESS) {
+            opened_session = FALSE;
+	    /* we handle this above */
+            did_setcred = FALSE;
 	    if (gdm_slave_should_complain ())
 		    gdm_error (_("Couldn't open session for %s"), login);
 	    goto pamerr;
@@ -678,11 +713,22 @@ authenticate_again:
 	    }
     }
 
+    did_setcred = FALSE;
+    opened_session = FALSE;
+
     if (pamh != NULL) {
+	    pam_handle_t *tmp_pamh;
+	    gdm_sigterm_block_push ();
+	    gdm_sigchld_block_push ();
+	    tmp_pamh = pamh;
+	    pamh = NULL;
+	    gdm_sigchld_block_pop ();
+	    gdm_sigterm_block_pop ();
+
 	    /* Throw away the credentials */
 	    if (credentials_set)
-		    pam_setcred (pamh, PAM_DELETE_CRED);
-	    pam_end (pamh, pamerr);
+		    pam_setcred (tmp_pamh, PAM_DELETE_CRED);
+	    pam_end (tmp_pamh, pamerr);
     }
     pamh = NULL;
     
@@ -831,17 +877,24 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display,
 	    goto setup_pamerr;
     }
 
+    did_setcred = TRUE;
+
     /* Set credentials */
     pamerr = pam_setcred (pamh, PAM_ESTABLISH_CRED);
     if (pamerr != PAM_SUCCESS) {
+	    did_setcred = FALSE;
 	    if (gdm_slave_should_complain ())
 		    gdm_error (_("Couldn't set credentials for %s"), login);
 	    goto setup_pamerr;
     }
 
+    opened_session = TRUE;
+
     /* Register the session */
     pamerr = pam_open_session (pamh, 0);
     if (pamerr != PAM_SUCCESS) {
+	    did_setcred = FALSE;
+	    opened_session = FALSE;
 	    /* Throw away the credentials */
 	    pam_setcred (pamh, PAM_DELETE_CRED);
 
@@ -850,6 +903,7 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display,
 	    goto setup_pamerr;
     }
 
+
     /* Workaround to avoid gdm messages being logged as PAM_pwdb */
     closelog ();
     openlog ("gdm", LOG_PID, LOG_DAEMON);
@@ -863,8 +917,20 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display,
     
  setup_pamerr:
     
-    if (pamh != NULL)
-	    pam_end (pamh, pamerr);
+    did_setcred = FALSE;
+    opened_session = FALSE;
+    if (pamh != NULL) {
+	    pam_handle_t *tmp_pamh;
+
+	    gdm_sigterm_block_push ();
+	    gdm_sigchld_block_push ();
+	    tmp_pamh = pamh;
+	    pamh = NULL;
+	    gdm_sigchld_block_pop ();
+	    gdm_sigterm_block_pop ();
+
+	    pam_end (tmp_pamh, pamerr);
+    }
     pamh = NULL;
     
     /* Workaround to avoid gdm messages being logged as PAM_pwdb */
@@ -894,16 +960,38 @@ gdm_verify_cleanup (GdmDisplay *d)
 
 	if (pamh != NULL) {
 		gint pamerr;
+		pam_handle_t *tmp_pamh;
+		gboolean old_opened_session;
+		gboolean old_did_setcred;
+
+		gdm_debug ("Running gdm_verify_cleanup and pamh != NULL");
+
+		gdm_sigterm_block_push ();
+		gdm_sigchld_block_push ();
+		tmp_pamh = pamh;
+		pamh = NULL;
+		old_opened_session = opened_session;
+		opened_session = FALSE;
+		old_did_setcred = did_setcred;
+		did_setcred = FALSE;
+		gdm_sigchld_block_pop ();
+		gdm_sigterm_block_pop ();
+
+		pamerr = PAM_SUCCESS;
 
 		/* Close the users session */
-		pamerr = pam_close_session (pamh, 0);
+		if (old_opened_session) {
+			gdm_debug ("Running pam_close_session");
+			pamerr = pam_close_session (tmp_pamh, 0);
+		}
 
 		/* Throw away the credentials */
-		pamerr = pam_setcred (pamh, PAM_DELETE_CRED);
+		if (old_did_setcred) {
+			gdm_debug ("Running pam_setcred with PAM_DELETE_CRED");
+			pamerr = pam_setcred (tmp_pamh, PAM_DELETE_CRED);
+		}
 
-		if (pamh != NULL)
-			pam_end (pamh, pamerr);
-		pamh = NULL;
+		pam_end (tmp_pamh, pamerr);
 
 		/* Workaround to avoid gdm messages being logged as PAM_pwdb */
 		closelog ();
@@ -916,6 +1004,9 @@ gdm_verify_cleanup (GdmDisplay *d)
 	setgroups (1, groups);
 
 	cur_gdm_disp = NULL;
+
+	/* reset limits */
+	gdm_reset_limits ();
 }
 
 
@@ -934,11 +1025,14 @@ gdm_verify_check (void)
 	pam_handle_t *ph = NULL;
 
 	if (pam_start ("gdm", NULL, &standalone_pamc, &ph) != PAM_SUCCESS) {
+		ph = NULL; /* be anal */
+
 		closelog ();
 		openlog ("gdm", LOG_PID, LOG_DAEMON);
 
-		gdm_text_message_dialog
-			(_("Can't find PAM configuration for gdm."));
+		if ( ! no_console)
+			gdm_text_message_dialog
+				(_("Can't find PAM configuration for gdm."));
 		gdm_fail ("gdm_verify_check: %s",
 			  _("Can't find PAM configuration for gdm."));
 	}
